import 'package:flutter_test/flutter_test.dart';
import 'package:lokalise_flutter_sdk/src/ota/data/repositories/local_bundle_repository.dart';
import 'package:lokalise_flutter_sdk/src/ota/data/repositories/remote_bundle_repository.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/dto/get_up_to_date_bundle_dto.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/exceptions/bundle_not_found_exception.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/exceptions/internal_exception.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/exceptions/lokalise_exception.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/models/bundle.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/models/credentials.dart';
import 'package:lokalise_flutter_sdk/src/ota/domain/use_cases/get_up_to_date_bundle_use_case.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';

import 'get_up_to_date_bundle_use_case_test.mocks.dart';

@GenerateMocks([RemoteBundleRepository, LocalBundleRepository])
void main() {
  late MockRemoteBundleRepository remoteRepo;
  late MockLocalBundleRepository localRepo;
  late GetUpToDateBundleUseCase useCase;

  setUp(() {
    remoteRepo = MockRemoteBundleRepository();
    localRepo = MockLocalBundleRepository();
    useCase = GetUpToDateBundleUseCase(
      localBundleRepository: localRepo,
      remoteBundleRepository: remoteRepo,
    );
  });

  group('GetUpToDateBundleUseCase - getBundle', () {
    test('throw unexpected exception', () async {
      // Given
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: 1,
        preRelease: true,
      );
      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenThrow(Exception());

      // When-Then
      expect(useCase.getBundle(dto: dto), throwsA(isA<LokaliseException>()));

      verifyNever(localRepo.saveBundle(bundle: anyNamed('bundle')));
    });

    test('throw InternalException', () async {
      // Given
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: 1,
        preRelease: true,
      );
      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenThrow(InternalException.unexpected(param: 'error'));

      // When-Then
      expect(useCase.getBundle(dto: dto), throwsA(isA<LokaliseException>()));

      verifyNever(localRepo.saveBundle(bundle: anyNamed('bundle')));
    });

    test('throw BundleNotFoundException', () async {
      // Given
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: 1,
        preRelease: true,
      );

      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenThrow(BundleNotFoundException());
      when(localRepo.removeBundle()).thenAnswer((_) => Future.value(true));

      // When
      final res = await useCase.getBundle(dto: dto);

      // Then
      expect(res.languageBundles, isEmpty);
      expect(res.projectId, equals(dto.credentials.projectId));
      expect(res.appVersion, dto.appVersion);
      expect(res.translationVersion, equals(0));

      verifyNever(localRepo.saveBundle(bundle: anyNamed('bundle')));
      verify(localRepo.removeBundle()).called(1);
    });

    test('remote repo returns null', () async {
      // Given
      const translationVersion = 123;
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: translationVersion,
        preRelease: true,
      );
      final localBundle = Bundle(
        projectId: 'local_test',
        translationVersion: translationVersion,
        appVersion: 'local_test',
        languageBundles: [],
      );

      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenAnswer((realInvocation) => Future.value(null));
      when(localRepo.getBundle()).thenReturn(localBundle);

      // When
      final res = await useCase.getBundle(dto: dto);

      // Then
      expect(res, equals(localBundle));

      verifyNever(localRepo.saveBundle(bundle: anyNamed('bundle')));
      verify(localRepo.getBundle()).called(1);
    });

    test('local bundle is the latest', () async {
      // Given
      const translationVersion = 123;
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: translationVersion,
        preRelease: true,
      );
      final bundle = Bundle(
        projectId: 'test',
        translationVersion: translationVersion,
        appVersion: 'test',
        languageBundles: [],
      );
      final localBundle = Bundle(
        projectId: 'local_test',
        translationVersion: translationVersion,
        appVersion: 'local_test',
        languageBundles: [],
      );

      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenAnswer((realInvocation) => Future.value(bundle));
      when(localRepo.getBundle()).thenReturn(localBundle);

      // When
      final res = await useCase.getBundle(dto: dto);

      // Then
      expect(res, equals(localBundle));

      verifyNever(localRepo.saveBundle(bundle: anyNamed('bundle')));
      verify(localRepo.getBundle()).called(1);
    });

    test('newer bundle received', () async {
      // Given
      const translationVersion = 123;
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: translationVersion,
        preRelease: true,
      );
      final bundle = Bundle(
        projectId: 'test',
        translationVersion: translationVersion + 1,
        appVersion: 'test',
        languageBundles: [],
      );

      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenAnswer((_) => Future.value(bundle));

      when(localRepo.saveBundle(bundle: bundle))
          .thenAnswer((_) => Future.value(true));

      // When
      final res = await useCase.getBundle(dto: dto);

      // Then
      expect(res, equals(bundle));

      verify(localRepo.saveBundle(bundle: bundle)).called(1);
    });

    test('older bundle received', () async {
      // Given
      const translationVersion = 123;
      final dto = GetUpToDateBundleDto(
        credentials: Credentials(projectId: 'test', token: 'test'),
        appVersion: 'test',
        translationVersion: translationVersion,
        preRelease: true,
      );
      final bundle = Bundle(
        projectId: 'test',
        translationVersion: translationVersion - 1,
        appVersion: 'test',
        languageBundles: [],
      );

      when(remoteRepo.getBundle(getUpToDateBundleDto: dto))
          .thenAnswer((_) => Future.value(bundle));

      when(localRepo.saveBundle(bundle: bundle))
          .thenAnswer((_) => Future.value(true));

      // When
      final res = await useCase.getBundle(dto: dto);

      // Then
      expect(res, equals(bundle));

      verify(localRepo.saveBundle(bundle: bundle)).called(1);
    });
  });
}
